"""The configurator for Wingman."""

from __future__ import annotations

import argparse
import types
from pathlib import Path
from typing import Any

import numpy as np
import wandb
import yaml

from wingman.exceptions import WingmanException


class LockedNamespace(types.SimpleNamespace):
    """A locked namespace, where new attributes can be set but old ones cannot be overwritten."""

    def __setattr__(self, name, value):
        """__setattr__."""
        if hasattr(self, name):
            raise WingmanException(
                "Can't override values for locked_namespace. "
                f"Tried to set `{name}` which already has value {value}."
            )
        super().__setattr__(name, value)


def dict_to_locked_ns(nested_dict: dict[str, Any]) -> LockedNamespace:
    """dict_to_ns.

    Args:
    ----
        nested_dict (dict[str, Any]): nested_dict

    Returns:
    -------
        types.SimpleNamespace:

    """
    return LockedNamespace(
        **{
            k: (dict_to_locked_ns(v) if isinstance(v, dict) else v)
            for k, v in nested_dict.items()
        }
    )


def check_dict_superset(base: dict[str, Any], target: dict[str, Any]) -> bool:
    """Checks that the `base` is a superset of the `target`.

    Args:
    ----
        base (dict[str, Any]): base
        target (dict[str, Any]): target

    Returns:
    -------
        bool:

    """
    for k, v in target.items():
        if k not in base:
            return False
        if isinstance(v, dict):
            check_dict_superset(base[k], v)
    return True


def dict_cli_overrides(config_dict: dict[str, Any]) -> dict[str, Any]:
    """When provided a `config_dict`, allows the overriding of values via the cli.

    For example, if we have the following dict:
    ```
    data = {
        "a": {
            "b": 3
        }
    }
    ```

    This function allows us to override the value at "b" via
    ```
    python3 <program.py> --a.b=5
    ```

    Args:
    ----
        config_dict (dict[str, Any]): a nested dictionary of values.

    Returns:
    -------
        dict[str, Any]: an altered nested dictionary.

    """

    def nested_argparse(
        parser: argparse.ArgumentParser,
        nested_dict: dict[str, Any],
        basename: str = "",
    ) -> argparse.ArgumentParser:
        """Builds an argparser by recursively nesting arguments.

        Args:
        ----
            parser (argparse.ArgumentParser): parser
            nested_dict (dict[str, Any]): nested_dict
            basename (str): basename

        Returns:
        -------
            argparse.ArgumentParser: the resulting argparser with nested arguments.

        """
        for k, v in nested_dict.items():
            # extend the basename
            name = f"{basename}.{k}" if basename else k

            # if dict, we need to recurse
            if isinstance(v, dict):
                parser = nested_argparse(
                    parser=parser,
                    nested_dict=v,
                    basename=name,
                )

            # otherwise, just add to parser like usual
            else:
                parser.add_argument(
                    f"--{name}",
                    type=type(v),
                    nargs="?",
                    const=True,
                    default=v,
                    help="None",
                )

        return parser

    # recursively convert arguments to cli args
    raw_overrides = vars(
        nested_argparse(
            parser=argparse.ArgumentParser(allow_abbrev=False),
            nested_dict=config_dict,
        ).parse_args()
    )

    # pull out each item in the raw overrides and
    # replace the variable within the main config
    for joint_k, v in raw_overrides.items():
        k_list = joint_k.split(sep=".")
        current = config_dict
        for key in k_list[:-1]:
            current = current[key]
        current[k_list[-1]] = v

    return config_dict


def generate_wingman_config(config_yaml: Path | str) -> LockedNamespace:
    """Reads the yaml file provided at init and converts it to commandline arguments."""
    # read in the file
    with open(config_yaml) as f:
        config_dict = yaml.load(f, Loader=yaml.FullLoader)

    # check the config is a superset of the expected yaml
    with open(Path(__file__).parent / "config.yaml") as f:
        if not check_dict_superset(
            base=config_dict,
            target=yaml.load(f, Loader=yaml.FullLoader),
        ):
            raise WingmanException(
                f"Incomplete yaml file `{config_yaml}` used. "
                "The yaml structure must be a superset of the yaml generated via `wingman-generate-config`."
            )

    # allow cli to override dict values
    config_dict = dict_cli_overrides(config_dict)

    # override model_id if needed
    if not config_dict["model"]["id"]:
        config_dict["model"]["id"] = str(np.random.randint(999999))

    # conditionally wandb
    if config_dict["wandb"]["enable"]:
        # generate the wandb run display name
        if config_dict["wandb"]["run"]["name"] != "":
            run_name = (
                f'{config_dict["wandb"]["run"]["name"]}, v={config_dict["model"]["id"]}'
            )
        else:
            run_name = config_dict["wandb"]["model"]["id"]

        # initialize wandb
        wandb.init(
            project=config_dict["wandb"]["project"]["name"],
            entity=config_dict["wandb"]["project"]["entity"],
            config=config_dict,
            name=run_name,
            notes=config_dict["wandb"]["run"]["notes"],
        )

        # optionally save code
        if config_dict["wandb"]["save_code"]:
            raise NotImplementedError("Save code is not yet implemented.")
            wandb.run.log_code(".", exclude_fn=lambda path: "venv" in path)

    return dict_to_locked_ns(config_dict)



"""For fancy printing in Wingman."""

from __future__ import annotations

import re
from pathlib import Path
from typing import Any, Literal

# colour list
c_colors = {
    "HEADER": "\033[95m",
    "OKBLUE": "\033[94m",
    "OKCYAN": "\033[96m",
    "OKGREEN": "\033[92m",
    "WARNING": "\033[93m",
    "FAIL": "\033[91m",
    "BOLD": "\033[1m",
    "UNDERLINE": "\033[4m",
}
end_c = "\033[0m"


def cstr(
    x: Any,
    ctype: Literal[
        "HEADER", "OKBLUE", "OKCYAN", "OKGREEN", "WARNING", "FAIL", "BOLD", "UNDERLINE"
    ],
) -> str:
    """Makes a string colourful.

    Args:
    ----
        x (Any): the string
        ctype (str): the colour

    Returns:
    -------
        str: the coloured string

    """
    return f"{c_colors[ctype]}{x}{end_c}"


log_flag = cstr(cstr("wingman", "BOLD"), "OKCYAN")
ansi_escape = re.compile(r"(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]")


def wm_print(x: Any, log_file: Path | str = ""):
    """Prints out strings decorated with the wingman status.

    Args:
    ----
        x (Any): the input string
        log_file (str): a target path to save the log file if any

    """
    print(f"{log_flag}: {x}")

    if not log_file:
        return

    with open(log_file, "a") as f:
        f.write(ansi_escape.sub("", x))
        f.write("\n")
        f.write("\n")



"""Tests the replay buffer module."""

from __future__ import annotations

from copy import deepcopy
from itertools import product
from pprint import pformat
from typing import Literal

import pytest
import torch
from utils import (
    are_equivalent,
    create_memory,
    create_shapes,
    element_to_bulk_dim_swap,
    generate_random_dict_data,
    generate_random_flat_data,
)

# define the test configurations
_random_rollovers = [True, False]
_modes = ["numpy", "torch"]
_devices = [torch.device("cpu")]
if torch.cuda.is_available():
    _devices.append(torch.device("cuda:0"))
_store_on_devices = [True, False]
_use_dict = [False, True]
ALL_CONFIGURATIONS = list(
    product(
        _random_rollovers,
        _modes,
        _devices,
        _store_on_devices,
        _use_dict,
    )
)


@pytest.mark.parametrize(
    "random_rollover, mode, device, store_on_device, use_dict",
    ALL_CONFIGURATIONS,
)
def test_bulk(
    random_rollover: bool,
    mode: Literal["numpy", "torch"],
    device: torch.device,
    store_on_device: bool,
    use_dict: bool,
):
    """Tests repeatedly bulking the buffer and whether it rollovers correctly."""
    bulk_size = 7
    mem_size = 11
    shapes = create_shapes(use_dict=use_dict, bulk_size=bulk_size)
    memory = create_memory(
        mem_size=mem_size,
        mode=mode,
        device=device,
        store_on_device=store_on_device,
        random_rollover=random_rollover,
        use_dict=use_dict,
    )

    for iteration in range(10):
        # try to stuff:
        # a) (bulk_size, 3) array
        # b) (bulk_size,) array
        data = []
        for shape in shapes:
            if isinstance(shape, (list, tuple)):
                data.append(
                    generate_random_flat_data(shape=(bulk_size, *shape), mode=mode)
                )
            elif isinstance(shape, dict):
                data.append(generate_random_dict_data(shapes=shape, mode=mode))
            else:
                raise ValueError
        memory.push(data, bulk=True)

        # reverse the data
        # on insertion we had [element_dim, bulk_dim, *data_shapes]
        # on comparison we want [bulk_dim, element_dim, *data_shapes]
        serialized_data = element_to_bulk_dim_swap(
            element_first_data=data,
            bulk_size=bulk_size,
        )

        # if random rollover and we're more than full, different matching technique
        if random_rollover and memory.is_full:
            num_matches = 0
            # match according to meshgrid
            for item1 in serialized_data:
                for item2 in memory:
                    num_matches += int(are_equivalent(item1, item2))

            assert (
                num_matches == bulk_size
            ), f"""Expected {bulk_size} matches inside the memory, got {num_matches}."""

            continue

        for step in range(bulk_size):
            item1 = serialized_data[step]
            item2 = memory[(iteration * bulk_size + step) % mem_size]
            assert are_equivalent(
                item1, item2
            ), f"""Something went wrong with rollover at iteration {iteration},
                step {step}, expected \n{pformat(item1)}, got \n{pformat(item2)}."""


@pytest.mark.parametrize(
    "random_rollover, mode, device, store_on_device, use_dict",
    ALL_CONFIGURATIONS,
)
def test_non_bulk(
    random_rollover: bool,
    mode: Literal["numpy", "torch"],
    device: torch.device,
    store_on_device: bool,
    use_dict: bool,
):
    """Tests the replay buffer generically."""
    mem_size = 11
    shapes = create_shapes(use_dict=use_dict)
    memory = create_memory(
        mem_size=mem_size,
        mode=mode,
        device=device,
        store_on_device=store_on_device,
        random_rollover=random_rollover,
        use_dict=use_dict,
    )

    previous_data = []
    for iteration in range(20):
        current_data = []
        for shape in shapes:
            if isinstance(shape, (list, tuple)):
                current_data.append(generate_random_flat_data(shape=shape, mode=mode))
            elif isinstance(shape, dict):
                current_data.append(generate_random_dict_data(shapes=shape, mode=mode))
            else:
                raise ValueError
        memory.push(current_data)

        # if random rollover and we're more than full, different matching method
        if random_rollover and memory.is_full:
            num_current_matches = 0
            num_previous_matches = 0
            for item in memory:
                num_current_matches += int(are_equivalent(item, current_data))
                num_previous_matches += int(are_equivalent(item, previous_data))

            assert (
                num_current_matches == 1
            ), f"""Expected 1 match for current_data, got {num_current_matches}."""
            assert (
                num_previous_matches <= 1
            ), f"""Expected 1 or 0 match for previous_data, got {num_previous_matches}."""

            continue

        # check the current data
        output = memory.__getitem__(iteration % mem_size)
        assert are_equivalent(
            output, current_data
        ), f"""Something went wrong with rollover at iteration {iteration},
            expected \n{pformat(current_data)}, got \n{pformat(output)}."""

        # check the previous data
        if iteration > 0:
            output = memory[(iteration - 1) % mem_size]
            assert are_equivalent(
                output, previous_data
            ), f"""Something went wrong with rollover at iteration {iteration},
                expected \n{pformat(previous_data)}, got \n{pformat(output)}."""

        previous_data = deepcopy(current_data)



"""Some simple Wingman functionality."""

from __future__ import annotations

from typing import Any

import numpy as np

try:
    import torch
except ImportError as e:
    raise ImportError(
        "Could not import torch, this is not bundled as part of Wingman and has to be installed manually."
    ) from e

__device = "cpu"
__device = "mps" if torch.backends.mps.is_available() else __device
__device = "cuda:0" if torch.cuda.is_available() else __device


def gpuize(
    input: np.ndarray | torch.Tensor,
    device: str | torch.device = __device,
    dtype: torch.dtype = torch.float32,
) -> torch.Tensor:
    """gpuize.

    Args:
    ----
        input (np.ndarray | torch.Tensor): the array that we want to gpuize
        device (str | torch.device): a string of the device we want to move the thing to
        dtype (torch.dtype): the datatype that the returned tensor should be

    """
    if torch.is_tensor(input):
        return input.to(device=device, dtype=dtype)  # pyright: ignore[reportAttributeAccessIssue]
    else:
        return torch.tensor(input, device=device, dtype=dtype)


def nested_gpuize(
    input: dict[str, Any],
    device: str | torch.device = __device,
    dtype: torch.dtype = torch.float32,
) -> dict[str, Any]:
    """Gpuize but for nested dictionaries of elements.

    Args:
    ----
        input (dict[str, Any]): the array that we want to gpuize
        device (str | torch.device): a string of the device we want to move the thing to
        dtype (torch.dtype): the datatype that the returned tensor should be

    """
    output = {}
    for key, value in input.items():
        if isinstance(value, dict):
            output[key] = nested_gpuize(value, device=device, dtype=dtype)
        else:
            output[key] = gpuize(value)
    return output


def cpuize(input: np.ndarray | torch.Tensor) -> np.ndarray:
    """cpuize.

    Args:
    ----
        input (np.ndarray | torch.Tensor): the array of the thing we want to put on the cpu

    """
    if torch.is_tensor(input):
        return input.detach().cpu().numpy()  # pyright: ignore[reportAttributeAccessIssue]
    else:
        return input  # pyright: ignore[reportReturnType]


def nested_cpuize(input: dict[str, Any]) -> dict[str, Any]:
    """Gpuize but for nested dictionaries of elements.

    Args:
    ----
        input (dict[str, Any]): the array of the thing we want to put on the cpu

    """
    output = {}
    for key, value in input.items():
        if isinstance(value, dict):
            output[key] = nested_cpuize(value)
        else:
            output[key] = cpuize(value)
    return output


def shutdown_handler(*_):
    """shutdown_handler.

    Args:
    ----
        _:

    """
    print("ctrl-c invoked")
    exit(0)



"""CLI functionality of Wingman."""

from __future__ import annotations

import glob
import os
import re
import shutil
import sys

from wingman.print_utils import cstr, wm_print


def _get_dir_size(path):
    total = 0
    with os.scandir(path) as it:
        for entry in it:
            if entry.is_file():
                total += entry.stat().st_size
            elif entry.is_dir():
                total += _get_dir_size(entry.path)
    return total


def compress_weights():
    """Compresses the weights directory by deleting empty directories and compressing all mark numbers down to 0.

    Example CLI usage:
    `wingman-compress-weights [optional weights directory]`
    """
    wm_print("---------------------------------------------")
    wm_print("Beginning compression...")
    wm_print("---------------------------------------------")

    target_dir = "./weights"
    if len(sys.argv) > 1:
        target_dir = sys.argv[1]

    # get the original filesize
    original_size = _get_dir_size(target_dir)

    # all ids
    id_dirs = glob.glob(f"./{target_dir}/*")

    # which folders to delete
    empty_ids = []

    # start going through each folder
    for id_dir in id_dirs:
        mark_dirs = [
            f
            for f in glob.glob(os.path.join(id_dir, "*"))
            if (re.compile(r"^-?\d+$").match(f.split("/")[-1]) and os.path.isdir(f))
        ]

        # if no weights files
        if len(mark_dirs) == 0:
            empty_ids.append(id_dir)
            continue

        # start finding the latest file
        _mark_number = 0
        latest_mark_dir = os.path.join(id_dir, f"{_mark_number}")
        while latest_mark_dir in mark_dirs:
            # as long as it exists, increment it
            _mark_number += 1
            latest_mark_dir = os.path.join(id_dir, f"{_mark_number}")

        # remove the latest mark number from the list, we want to keep this
        _mark_number = min(0, _mark_number - 1)
        latest_mark_dir = os.path.join(id_dir, f"{_mark_number}")
        if latest_mark_dir in mark_dirs:
            mark_dirs.remove(latest_mark_dir)

        # remove the intermediary from our list if it's there
        intermediary = os.path.join(id_dir, "-1")
        if intermediary in mark_dirs:
            mark_dirs.remove(intermediary)

        # exit if nothing to delete
        if len(mark_dirs) == 0:
            wm_print(f"Nothing to compress in {cstr(id_dir, 'OKGREEN')}.")
            continue

        # for the remainder of the list, delete all the weights
        wm_print(f"Compressing {id_dir}...")
        for mark_dir in mark_dirs:
            shutil.rmtree(mark_dir)

        # rename the latest weights to 0
        os.rename(latest_mark_dir, os.path.join(id_dir, "0"))

        wm_print(f"Compressed {cstr(id_dir, 'WARNING')}.")

    # delete flagged directories
    for id_dir in empty_ids:
        shutil.rmtree(id_dir)
        wm_print(f"Deleted {cstr(id_dir, 'WARNING')}.")

    # get the final filesize
    final_size = _get_dir_size(target_dir)

    # printout
    wm_print("---------------------------------------------")
    wm_print(f"Original disk usage: {cstr(original_size / 1e9, 'OKGREEN')} gigabytes.")
    wm_print(
        f"Removed {cstr((original_size - final_size) / 1e9, 'WARNING')} gigabytes."
    )
    wm_print(f"Final disk usage: {cstr(final_size / 1e9, 'WARNING')} gigabytes.")
    wm_print("---------------------------------------------")


def generate_yaml():
    """Generates the basic yaml file for the Wingman module.

    Example CLI usage:
    `wingman-generate-yaml [optional filename]`
    """
    source = os.path.join(os.path.dirname(__file__), "./config.yaml")

    target = "./config.yaml"
    if len(sys.argv) > 1:
        target = sys.argv[1]

    shutil.copyfile(source, target)

    wm_print(f"Generated config file at: {cstr(target, 'OKGREEN')}.")



"""Helper for converting a list of nested dictionaries to a nested dictionary of lists."""

from __future__ import annotations

import functools
from typing import Any, Generator

import numpy as np
import torch


def _iter_nested_keys(base_dict: dict[str, Any]) -> Generator[list[str], None, None]:
    """Given a nested dictionary, yields a list of keys to each element.

    Args:
    ----
        base_dict (dict[str, Any]): base_dict

    Returns:
    -------
        Generator[list[str], None, None]:

    """
    for key, value in base_dict.items():
        if isinstance(value, dict):
            for sub_keys in _iter_nested_keys(value):
                yield [key, *sub_keys]
        else:
            yield [key]


def listed_dict_to_dicted_list(
    list_dict: list[dict[str, Any]], stack: bool
) -> dict[str, Any]:
    """Given a list of nested dicts, returns a nested dict of lists.

    Args:
    ----
        list_dict (list[dict[str, Any]]): list_dict
        stack (bool): stack

    Returns:
    -------
        dict[str, Any]:

    """
    result = {}
    for key_list in _iter_nested_keys(list_dict[0]):
        # for each element in the expected dictionary
        # scaffold to that point
        ptr = result
        for key in key_list[:-1]:
            ptr = ptr.setdefault(key, {})

        # this goes through the main list of dicts
        # and collects elements at the position determined by the key_list
        dicted_list: list = [
            functools.reduce(lambda d, k: d[k], key_list, dict_item)
            for dict_item in list_dict
        ]

        # we can't call `concatenate` on a list of non-np.ndarray or non-torch.Tensor items
        if isinstance(dicted_list[0], np.ndarray) and len(dicted_list[0].shape) > 0:
            if stack:
                ptr[key_list[-1]] = np.stack(dicted_list, axis=0)
            else:
                ptr[key_list[-1]] = np.concatenate(dicted_list, axis=0)
        elif isinstance(dicted_list[0], torch.Tensor):
            if stack:
                ptr[key_list[-1]] = torch.stack(dicted_list, dim=0)
            else:
                ptr[key_list[-1]] = torch.concatenate(dicted_list, dim=0)
        else:
            if stack:
                ptr[key_list[-1]] = np.expand_dims(
                    np.asarray(dicted_list),
                    axis=0,
                )
            else:
                ptr[key_list[-1]] = np.asarray(dicted_list)
    return result



"""Replay buffer implementation with push, automatic overflow, and automatic torch dataset functionality."""

from __future__ import annotations

from enum import Enum
from typing import Literal, Sequence

import numpy as np

from wingman.exceptions import ReplayBufferException
from wingman.replay_buffer.core import ReplayBuffer

try:
    import torch
except ImportError as e:
    raise ImportError(
        "Could not import torch, this is not bundled as part of Wingman and has to be installed manually."
    ) from e

from wingman.print_utils import cstr, wm_print


class _Mode(Enum):
    """_Mode."""

    TORCH = 1
    NUMPY = 2


class FlatReplayBuffer(ReplayBuffer):
    """Replay Buffer implementation of a Torch or Numpy dataset."""

    def __init__(
        self,
        mem_size: int,
        mode: Literal["numpy", "torch"] = "numpy",
        device: torch.device = torch.device("cpu"),
        store_on_device: bool = False,
        random_rollover: bool = False,
    ):
        """__init__.

        Args:
        ----
            mem_size (int): number of transitions the replay buffer aims to hold
            mode (Literal["numpy", "torch"]): Whether to store data as "torch" or "numpy".
            device (torch.device): The target device that data will be retrieved to if "torch".
            store_on_device (bool): Whether to store the entire replay on the specified device, otherwise stored on CPU.
            random_rollover (bool): whether to rollover the data in the replay buffer once full or to randomly insert

        """
        super().__init__(mem_size=mem_size)

        # store the device
        self.device = device
        self.storage_device = self.device if store_on_device else torch.device("cpu")

        # random rollover
        self.random_rollover = random_rollover

        # store the mode
        if mode == "numpy":
            self.mode = _Mode.NUMPY
            self.mode_type = np.ndarray
            self.mode_caller = np
            self.mode_dtype = np.float32
        elif mode == "torch":
            self.mode = _Mode.TORCH
            self.mode_type = torch.Tensor
            self.mode_caller = torch
            self.mode_dtype = torch.float32
        else:
            raise ReplayBufferException(
                f"Unknown mode {mode}. Only `'numpy'` and `'torch'` are allowed."
            )

    def __getitem__(self, idx: int) -> list[np.ndarray | torch.Tensor]:
        """__getitem__.

        Args:
        ----
            idx (int): idx

        Returns:
        -------
            list[np.ndarray | torch.Tensor]:

        """
        return list(d[idx] for d in self.memory)

    def _format_data(
        self, thing: np.ndarray | torch.Tensor | float | int | bool, bulk: bool
    ) -> np.ndarray | torch.Tensor:
        """_format_data.

        Args:
        ----
            thing (np.ndarray | torch.Tensor | float | int | bool): thing
            bulk (bool): bulk

        Returns:
        -------
            np.ndarray | torch.Tensor:

        """
        if self.mode == _Mode.NUMPY:
            # cast to the right dtype
            data = np.asarray(
                thing,
                dtype=self.mode_dtype,  # pyright: ignore[reportArgumentType, reportCallIssue]
            )

            # dim check
            if bulk and len(data.shape) < 1:
                data = np.expand_dims(data, axis=-1)
        elif self.mode == _Mode.TORCH:
            # cast to the right dtype, store on CPU intentionally
            data = torch.asarray(
                thing,
                device=self.storage_device,
                dtype=self.mode_dtype,  # pyright: ignore[reportArgumentType]
            )
            data.requires_grad_(False)

            # dim check
            if bulk and len(data.shape) < 1:
                data = data.unsqueeze(-1)
        else:
            raise ReplayBufferException(
                f"Unknown mode {self.mode}. Only `'numpy'` and `'torch'` are allowed."
            )

        return data

    def push(
        self,
        data: Sequence[torch.Tensor | np.ndarray | float | int | bool],
        bulk: bool = False,
    ) -> None:
        """Adds transition tuples into the replay buffer.

        The data must be either:
        - an n-long tuple of a single transition
        - an n-long tuple of m transitions, ie: a list of [m, ...] np arrays with the `bulk` flag set to True

        Args:
        ----
            data (Sequence[torch.Tensor | np.ndarray | float | int | bool]): data
            bulk (bool): whether to bulk add stuff into the replay buffer

        """
        # cast to dtype and conditionally add batch dim
        array_data = [self._format_data(item, bulk=True) for item in data]

        # if nothing in the array, we can safely return
        # this can occur for example, when we do `memory.push([np.array([])])`
        if all([len(item) == 0 for item in array_data]):
            return

        if not bulk:
            bulk_size = 1
        else:
            bulk_size = data[0].shape[0]  # pyright: ignore
            # assert all items have same length
            if not all([len(item) == bulk_size for item in array_data]):  # pyright: ignore[reportArgumentType]
                raise ReplayBufferException(
                    "All things in data must have same len for the first dimension for bulk data. "
                    f"Received data with {[len(item) for item in array_data]} items respectively.",
                )

            # assert on memory lengths
            if self.mem_size < bulk_size:
                raise ReplayBufferException(
                    f"Bulk size ({bulk_size}) should be less than or equal to memory size ({self.mem_size}).",
                )

        # instantiate the memory if it does not exist
        if self.count == 0:
            self.memory = []
            if not bulk:
                self.memory.extend(
                    [
                        self.mode_caller.zeros(
                            (self.mem_size, *item.shape),
                            dtype=self.mode_dtype,  # pyright: ignore[reportArgumentType, reportCallIssue]
                        )
                        for item in array_data
                    ]
                )
            else:
                self.memory.extend(
                    [
                        self.mode_caller.zeros(
                            (self.mem_size, *item.shape[1:]),
                            dtype=self.mode_dtype,  # pyright: ignore[reportArgumentType, reportCallIssue]
                        )
                        for item in array_data
                    ]
                )

            # move everything to the storage device if torch
            if self.mode == _Mode.TORCH:
                self.memory = [array.to(self.storage_device) for array in self.memory]

            mem_size_bytes = sum([d.nbytes for d in self.memory])
            wm_print(
                cstr(f"Replay Buffer Size: {mem_size_bytes / 1e9} gigabytes.", "OKCYAN")
            )

        # assert that the number of lists in memory is same as data to push
        if len(array_data) != len(self.memory):
            raise ReplayBufferException(
                f"Data length not similar to memory buffer length. Replay buffer has {len(self.memory)} items. "
                f"But received {len(array_data)} items.",
            )

        # indexing for memory positions
        start = self.count % self.mem_size
        stop = min(start + bulk_size, self.mem_size)
        rollover = -(self.mem_size - start - bulk_size)
        if self.random_rollover:
            if not self.is_full:
                idx_front = np.arange(start, stop)
                idx_back = np.random.choice(
                    start,
                    size=np.maximum(rollover, 0),
                    replace=False,
                )
            else:
                idx_front = np.random.choice(
                    self.mem_size,
                    size=bulk_size,
                    replace=False,
                )
                idx_back = np.array([], dtype=np.int64)
        else:
            idx_front = np.arange(start, stop)
            idx_back = np.arange(0, rollover)
        idx = np.concatenate((idx_front, idx_back), axis=0)

        # put things in memory
        for memory, item in zip(self.memory, array_data):
            memory[idx] = item

        self.count += bulk_size

    def sample(self, batch_size: int) -> list[np.ndarray | torch.Tensor]:
        """sample.

        Args:
        ----
            batch_size (int): batch_size

        Returns:
        -------
            list[np.ndarray | torch.Tensor]:

        """
        idx = np.random.randint(
            0,
            len(self),
            size=np.minimum(len(self), batch_size),
        )
        if self.mode == _Mode.TORCH:
            return [item[idx].to(self.device) for item in self.memory]
        else:
            return [item[idx] for item in self.memory]



"""The main Wingman package."""

# ruff: noqa: F401
from .wingman import Wingman as Wingman

try:
    import torch as _

    from .neural_blocks import NeuralBlocks as NeuralBlocks
except ImportError:
    import warnings

    warnings.warn(
        "Could not import torch, "
        "this is not bundled as part of Wingman and has to be installed manually, "
        "as a result, `NeuralBlocks` and `ReplayBuffer` are unavailable.",
        category=RuntimeWarning,
    )



"""Exceptions."""

from wingman.print_utils import cstr


class WingmanException(Exception):
    """WingmanException."""

    def __init__(self, message: str = ""):
        """__init__.

        Args:
        ----
            message (str): the message

        """
        message = cstr(message, "FAIL")
        super().__init__(message)
        self.message = message


class NeuralBlocksException(Exception):
    """NeuralBlocksException."""

    def __init__(self, message: str = ""):
        """__init__.

        Args:
        ----
            message (str): the message

        """
        message = cstr(message, "FAIL")
        super().__init__(message)
        self.message = message


class ReplayBufferException(Exception):
    """ReplayBufferException."""

    def __init__(self, message: str = ""):
        """__init__.

        Args:
        ----
            message (str): the message

        """
        message = cstr(message, "FAIL")
        super().__init__(message)
        self.message = message



"""Tests the replay buffer utilities."""

from __future__ import annotations

import numpy as np
import pytest
from utils import are_equivalent

from wingman.replay_buffer.utils import listed_dict_to_dicted_list


@pytest.mark.parametrize(
    "stack",
    [True, False],
)
def test_dicted_list_to_listed_dict(stack: bool) -> None:
    """Tests the dicted_list_to_listed_dict function."""
    # the input
    listed_dict = [
        {"a": {"x": 1, "y": 2}, "b": [3, 4]},
        {"a": {"x": 5, "y": 6}, "b": [7, 8]},
        {"a": {"x": 9, "y": 10}, "b": [11, 12]},
    ]

    # the target output
    first = np.asarray([1, 5, 9])
    second = np.asarray([2, 6, 10])
    third = np.asarray(
        [
            [3, 4],
            [7, 8],
            [11, 12],
        ]
    )
    if stack:
        first = np.expand_dims(first, axis=0)
        second = np.expand_dims(second, axis=0)
        third = np.expand_dims(third, axis=0)
    target_dicted_list = {
        "a": {
            "x": first,
            "y": second,
        },
        "b": third,
    }

    # convert and check
    created_dicted_list = listed_dict_to_dicted_list(listed_dict, stack=stack)

    assert are_equivalent(
        target_dicted_list, created_dicted_list
    ), f"Expected {target_dicted_list=} to equal {created_dicted_list=}."



"""Base replay buffer class."""

from __future__ import annotations

import warnings
from abc import abstractmethod
from typing import Any, Generator, Sequence

from prefetch_generator import prefetch


class ReplayBuffer:
    """Base replay buffer class."""

    def __init__(self, mem_size: int):
        """__init__.

        Args:
        ----
            mem_size (int): mem_size

        """
        self.mem_size = mem_size
        self.count = 0
        self.memory = []

    def __len__(self) -> int:
        """The number of memory items this replay buffer is holding.

        Returns
        -------
            int:

        """
        return min(self.mem_size, self.count)

    def __repr__(self) -> str:
        """Printouts parameters of this replay buffer.

        Returns
        -------
            str:

        """
        return f"""ReplayBuffer of size {self.mem_size} with {len(self.memory)} elements. \n
        A brief view of the memory: \n
        {self.memory}
        """

    @property
    def is_full(self) -> bool:
        """Whether or not the replay buffer has reached capacity.

        Returns
        -------
            bool: whether the buffer is full

        """
        return self.count >= self.mem_size

    def merge(self, other: ReplayBuffer) -> None:
        """Merges another replay buffer into this replay buffer via the `push` method.

        Args:
        ----
            other (ReplayBuffer): other

        Returns:
        -------
            None:

        """
        self.push(
            [m[: min(other.mem_size, other.count)] for m in other.memory],
            bulk=True,
        )

    @prefetch(max_prefetch=1)
    def iter_sample(
        self, batch_size: int, num_iter: int
    ) -> Generator[Sequence[Any], None, None]:
        """iter_sample.

        Args:
        ----
            batch_size (int): batch_size
            num_iter (int): num_iter

        Returns:
        -------
            Generator[Sequence[np.ndarray | torch.Tensor], None, None]:

        """
        for _ in range(num_iter):
            yield (self.sample(batch_size=batch_size))

    @abstractmethod
    def sample(self, batch_size: int) -> Sequence[Any]:
        """sample.

        Args:
        ----
            batch_size (int): batch_size

        Returns:
        -------
            Sequence[Any]:

        """
        raise NotImplementedError

    @abstractmethod
    def push(
        self,
        data: Sequence[Any],
        bulk: bool = False,
    ) -> None:
        """push.

        Args:
        ----
            data (Sequence[Any]): data
            bulk (bool): bulk

        Returns:
        -------
            None:

        """
        raise NotImplementedError

    @abstractmethod
    def __getitem__(self, idx: int) -> Sequence[Any]:
        """__getitem__.

        Args:
        ----
            idx (int): idx

        Returns:
        -------
            Sequence[Any]:

        """
        raise NotImplementedError


class ReplayBufferWrapper(ReplayBuffer):
    """ReplayBufferWrapper."""

    def __init__(self, base_buffer: ReplayBuffer):
        """__init__.

        Args:
        ----
            base_buffer (ReplayBuffer): base_buffer

        """
        self.base_buffer = base_buffer
        self.mem_size = base_buffer.mem_size

    @property
    def count(self) -> int:
        """The number of transitions that's been through this buffer."""
        return self.base_buffer.count

    @property
    def memory(self) -> list[Any]:
        """The core memory of this buffer."""
        warnings.warn(
            "Accessing the core of `ReplayBufferWrapper` returns the "
            "memory of the base buffer, not the wrapped buffer",
            category=RuntimeWarning,
        )
        return self.base_buffer.memory

    def merge(self, other: ReplayBuffer) -> None:
        """Merges another replay buffer into this replay buffer via the `push` method.

        Args:
        ----
            other (ReplayBuffer): other

        Returns:
        -------
            None:

        """
        # can't merge when the other one has 0 items
        assert other.count > 0

        # if no count, we need to manually push one item first to build the index
        if self.count == 0:
            self.push(other[0])

        # if the other only has one item, our merge is done
        if other.count == 1:
            return

        self.base_buffer.push(
            [m[1 : min(other.mem_size, other.count)] for m in other.memory],
            bulk=True,
        )

    def __len__(self) -> int:
        """The number of memory items this replay buffer is holding."""
        return len(self.base_buffer)

    def __repr__(self) -> str:
        """Printouts parameters of this replay buffer."""
        return f"""ReplayBuffer of size {self.mem_size} with {len(self.memory)} elements. \n
        A brief view of the memory: \n
        {self.base_buffer}
        """

    def __getitem__(self, idx: int) -> Sequence[Any]:
        """__getitem__.

        Args:
        ----
            idx (int): idx

        Returns:
        -------
            Sequence[Any]:

        """
        return self.wrap_data(self.base_buffer[idx])

    def push(
        self,
        data: Sequence[Any],
        bulk: bool = False,
    ) -> None:
        """push.

        Args:
        ----
            data (Sequence[Any]): data
            bulk (bool): bulk

        Returns:
        -------
            None:

        """
        self.base_buffer.push(
            data=self.unwrap_data(
                wrapped_data=data,
                bulk=bulk,
            ),
            bulk=bulk,
        )

    def sample(self, batch_size: int) -> Sequence[Any]:
        """sample.

        Args:
        ----
            batch_size (int): batch_size

        Returns:
        -------
            Sequence[Any]:

        """
        return self.wrap_data(self.base_buffer.sample(batch_size=batch_size))

    @abstractmethod
    def unwrap_data(self, wrapped_data: Sequence[Any], bulk: bool) -> Sequence[Any]:
        """Unwraps data from the underlying data into an unwrapped format.

        This is called when packing the data into the `base_buffer`.

        Args:
        ----
            wrapped_data (Sequence[Any]): wrapped_data
            bulk (bool): bulk

        Returns:
        -------
            Sequence[Any]:

        """
        raise NotImplementedError

    @abstractmethod
    def wrap_data(self, unwrapped_data: Sequence[Any]) -> Sequence[Any]:
        """Wraps data from the underlying data into a wrapped format.

        This is called when sampling data from `base_buffer`.

        Args:
        ----
            unwrapped_data (Sequence[Any]): unwrapped_data

        Returns:
        -------
            Sequence[Any]:

        """
        raise NotImplementedError



"""Wrapper to convert a FlatReplayBuffer into one that accepts nested dicts."""

from __future__ import annotations

from typing import Any, Mapping, Sequence, Union

import numpy as np

from wingman.exceptions import ReplayBufferException
from wingman.replay_buffer.core import ReplayBufferWrapper
from wingman.replay_buffer.flat_replay_buffer import FlatReplayBuffer

try:
    import torch
except ImportError as e:
    raise ImportError(
        "Could not import torch, this is not bundled as part of Wingman and has to be installed manually."
    ) from e

_NestedDict = Mapping[str, Union[int, "_NestedDict"]]


class DictReplayBufferWrapper(ReplayBufferWrapper):
    """Replay Buffer Wrapper that allows the underlying replay buffer to take in nested dicts."""

    def __init__(self, replay_buffer: FlatReplayBuffer) -> None:
        """__init__.

        If bulk adding items, this expects dictionary items to be a dictionary of lists, NOT a list of dictionaries.

        Args:
        ----
            self:
            replay_buffer (FlatReplayBuffer): replay_buffer

        Returns:
        -------
            None:

        """
        super().__init__(replay_buffer)

        # this is a list where:
        # - if the element is an integer, specifies the location in the unwrapped flat data that the wrapped element in this index should be
        # - if it's a dictionary, specifies that the data at this index in the wrapped data should be a dictionary, where
        #   - if the value is an int, is the location of this data in the flat dict
        #   - if the value is a dict, means that this data holds a nested dict
        self.mapping: list[int | _NestedDict] = []
        self.total_elements = 0

    @staticmethod
    def _recursive_unpack_dict_mapping(
        data_dict: dict[str, np.ndarray | torch.Tensor],
        start_idx: int,
    ) -> tuple[_NestedDict, int]:
        """Recursively unpacks a dictionary into a mapping.

        Args:
        ----
            data_dict (dict[str, np.ndarray | torch.Tensor]): data_dict
            start_idx (int): start_idx

        Returns:
        -------
            tuple[_NestedDict, int]:

        """
        mapping: _NestedDict = dict()
        idx = start_idx

        for key, value in data_dict.items():
            if isinstance(value, dict):
                mapping[key], idx = (
                    DictReplayBufferWrapper._recursive_unpack_dict_mapping(
                        value, start_idx=idx
                    )
                )
            else:
                mapping[key] = idx
                idx += 1

        return mapping, idx

    @staticmethod
    def _generate_mapping(
        wrapped_data: Sequence[
            dict[str, np.ndarray | torch.Tensor]
            | np.ndarray
            | torch.Tensor
            | float
            | int
            | bool
        ],
    ) -> tuple[list[int | _NestedDict], int]:
        """Generates a mapping from wrapped data.

        For example:
            data = [
                32,
                65,
                {
                    "a": 5,
                    "b": {
                        "c": 6,
                        "d": 7,
                        "e": {
                            "f": 8
                        }
                    },
                    "g": {
                        "h": 9
                    }
                },
                100,
            ]

            becomes:

            [
                0,
                1,
                {
                    'a': 2,
                    'b': {
                        'c': 3,
                        'd': 4,
                        'e': {'f': 5}
                    },
                    'g': {'h': 6}
                },
                7
            ]

        Args:
        ----
            wrapped_data (Sequence[dict[str, np.ndarray | torch.Tensor] | np.ndarray | torch.Tensor | float | int | bool]): wrapped_data

        Returns:
        -------
            tuple[list[int | _NestedDict], int]:

        """
        mapping: list[int | _NestedDict] = []
        idx = 0

        for item in wrapped_data:
            if isinstance(item, dict):
                dict_mapping, idx = (
                    DictReplayBufferWrapper._recursive_unpack_dict_mapping(
                        item, start_idx=idx
                    )
                )
                mapping.append(dict_mapping)
            else:
                mapping.append(idx)
                idx += 1

        return mapping, idx

    @staticmethod
    def _recursive_unpack_dict_data(
        data_dict: dict[str, np.ndarray | torch.Tensor],
        unwrapped_data_target: list[Any],
        mapping: _NestedDict,
    ) -> list[Any]:
        """Recursively unpacks dictionary data into a sequence of items that FlatReplayBuffer can use.

        Args:
        ----
            data_dict (dict[str, np.ndarray | torch.Tensor]): data_dict
            unwrapped_data_target (list[Any]): unwrapped_data_target
            mapping (_NestedDict): mapping

        Returns:
        -------
            list[Any]:

        """
        for key, value in data_dict.items():
            if isinstance((idx_map := mapping[key]), int):
                unwrapped_data_target[idx_map] = value

            elif isinstance((idx_map := mapping[key]), dict):
                if not isinstance(value, dict):
                    raise ReplayBufferException(
                        "Something went wrong with data unwrapping.\n"
                        f"Expected a dictionary for key {key} within the data, but got {type(value)}."
                    )

                unwrapped_data_target = (
                    DictReplayBufferWrapper._recursive_unpack_dict_data(
                        data_dict=value,
                        unwrapped_data_target=unwrapped_data_target,
                        mapping=idx_map,
                    )
                )
            else:
                raise ValueError("Not supposed to be here")

        return unwrapped_data_target

    def unwrap_data(
        self,
        wrapped_data: Sequence[
            dict[str, Any] | np.ndarray | torch.Tensor | float | int | bool
        ],
        bulk: bool,
    ) -> Sequence[np.ndarray | torch.Tensor | float | int | bool]:
        """Unwraps dictionary data into a sequence of items that FlatReplayBuffer can use.

        If bulk adding items, this expects dictionary items to be a dictionary of lists, NOT a list of dictionaries.

        Args:
        ----
            wrapped_data (Sequence[dict[str, Any] | np.ndarray | torch.Tensor | float | int | bool]): wrapped_data
            bulk (bool): bulk

        Returns:
        -------
            Sequence[np.ndarray | torch.Tensor | float | int | bool]:

        """
        if len(self) == 0:
            self.mapping, self.total_elements = self._generate_mapping(
                wrapped_data=wrapped_data
            )

        if len(self.mapping) != len(wrapped_data):
            raise ReplayBufferException(
                "Something went wrong with data unwrapping.\n"
                f"Expected `wrapped_data` to have {len(self.mapping)} items, but got {len(wrapped_data)}."
            )

        # holder for the unwrapped data
        unwrapped_data: list[Any] = [None] * self.total_elements

        for i, (mapping, data_item) in enumerate(zip(self.mapping, wrapped_data)):
            # if the mapping says it's an int, then just set the data without nesting
            if isinstance(mapping, int):
                unwrapped_data[mapping] = data_item

            # if it's a dict, then we need to recursively unpack
            elif isinstance(mapping, dict):
                if not isinstance(data_item, dict):
                    raise ReplayBufferException(
                        "Something went wrong with data unwrapping.\n"
                        f"Expected `wrapped_data` at element {i} to be a dict, but got {type(data_item)}."
                    )

                unwrapped_data = self._recursive_unpack_dict_data(
                    data_dict=data_item,
                    unwrapped_data_target=unwrapped_data,
                    mapping=mapping,
                )

        return unwrapped_data

    @staticmethod
    def _recursive_pack_dict_data(
        unwrapped_data: Sequence[Any],
        mapping: _NestedDict,
    ) -> dict[str, Any]:
        """Packs back a sequence of items into a dictionary structure according to a mapping.

        Args:
        ----
            unwrapped_data (Sequence[Any]): unwrapped_data
            mapping (_NestedDict): mapping

        Returns:
        -------
            dict[str, Any]:

        """
        data_dict = dict()
        for key, idx_map in mapping.items():
            if isinstance(idx_map, int):
                data_dict[key] = unwrapped_data[idx_map]
            elif isinstance(idx_map, dict):
                data_dict[key] = DictReplayBufferWrapper._recursive_pack_dict_data(
                    unwrapped_data=unwrapped_data,
                    mapping=idx_map,
                )
            else:
                raise ValueError("Not supposed to be here")

        return data_dict

    def wrap_data(
        self, unwrapped_data: Sequence[np.ndarray | torch.Tensor | float | int | bool]
    ) -> Sequence[
        dict[str, np.ndarray | torch.Tensor]
        | np.ndarray
        | torch.Tensor
        | float
        | int
        | bool
    ]:
        """Converts a sequence of items into a dictionary structure that is similar to that used during `push`.

        Args:
        ----
            self:
            unwrapped_data (Sequence[np.ndarray | torch.Tensor | float | int | bool]): unwrapped_data

        Returns:
        -------
            Sequence[dict[str, np.ndarray | torch.Tensor] | np.ndarray | torch.Tensor | float | int | bool]:

        """
        wrapped_data: list[Any] = [None] * len(self.mapping)

        for i, idx_map in enumerate(self.mapping):
            if isinstance(idx_map, int):
                wrapped_data[i] = unwrapped_data[idx_map]
            elif isinstance(idx_map, dict):
                wrapped_data[i] = DictReplayBufferWrapper._recursive_pack_dict_data(
                    unwrapped_data=unwrapped_data,
                    mapping=idx_map,
                )
            else:
                raise ValueError("Not supposed to be here")

        return wrapped_data



"""The core of Wingman."""

from __future__ import annotations

import math
import shutil
import time
from functools import cached_property
from pathlib import Path

import numpy as np
import wandb

from wingman.config_utils import LockedNamespace, generate_wingman_config
from wingman.exceptions import WingmanException
from wingman.print_utils import cstr, wm_print


class Wingman:
    """Wingman.

    Class to handle checkpointing of loss and handling of weights files.
    Minimal example:

    ```py
    helper = Wingman(./config.yaml)
    cfg = helper.cfg

    # load the model and optimizer
    model = Model().to(cfg.device)
    optim = optimizer.AdamW(model.parameters(), lr=cfg.learning_rate, amsgrad=True)

    # get the weight files if they exist
    have_file, model_dir, ckpt_dir = self.get_weight_files()
    if have_file:
        model.load(f"{ckpt_dir}/weights.pth")
        optim.load(f"{model_dir}/optim.pth")

    # run some training:
    while(training):
        ...
        # training code here
        ...

        update_weights, model_dir, ckpt_dir = self.checkpoint(loss, step_number)
        if update_weights:
            model.save(f"{ckpt_dir}/weights.pth")
            optim.save(f"{model_dir}/{ckpt_subdir}/optim.pth")
    ```
    """

    def __init__(
        self,
        config_yaml: str | Path,
    ):
        """__init__.

        Args:
        ----
            config_yaml (str): location of where the config yaml is described

        """
        # save our experiment description
        self.cfg: LockedNamespace = generate_wingman_config(config_yaml)

        # make sure that logging_interval is positive
        if self.cfg.logging.interval <= 0:
            raise WingmanException(
                cstr(
                    f"logging_interval must be a positive number, got {self.cfg.logging.interval}.",
                    "FAIL",
                )
            )

        # the logger
        self.log = dict()

        # runtime variables
        self._num_losses: int = 0
        self._cumulative_loss: float = 0.0
        self._lowest_cumulative_lost: float = math.inf
        self._next_log_step: int = self.cfg.logging.interval
        self._previous_ckpt_step: int = 0
        self._skips: int = 0

        # weight file variables
        self._current_ckpt: int = self.cfg.model.ckpt
        self._previous_ckpt: int = -1

        # file paths
        self._model_dir: Path = Path(self.cfg.model.save_directory) / str(
            self.cfg.model.id
        )
        self._log_file: Path = self._model_dir / "log.txt"
        self._lowest_loss_file: Path = self._model_dir / "lowest_loss.npy"

        wm_print("--------------𓆩𓆪--------------")
        wm_print(f"Using device {cstr(self.device, 'HEADER')}")
        wm_print(f"Saving weights to {cstr(self._model_dir, 'HEADER')}...")

        # check to record that we're in a new training session and save the config file
        if self._model_dir.is_dir():
            self._fresh_directory = False
        else:
            self._fresh_directory = True
            wm_print(
                cstr(
                    "New training instance detected, generating model directory in 3 seconds...",
                    "WARNING",
                ),
            )
            time.sleep(3)
            self._model_dir.mkdir(parents=True, exist_ok=True)
            shutil.copyfile(
                config_yaml,
                self._model_dir / "config_copy.yaml",
            )

    @property
    def _ckpt_dir(self) -> Path:
        ckpt_dir = self._model_dir / f"{self._current_ckpt}/"
        ckpt_dir.mkdir(parents=True, exist_ok=True)
        return ckpt_dir

    @cached_property
    def _temp_dir(self) -> Path:
        temp_dir = self._model_dir / "-1/"
        temp_dir.mkdir(parents=True, exist_ok=True)
        return temp_dir

    @cached_property
    def device(self):
        """__get_device."""
        try:
            import torch

            if torch.cuda.is_available():
                device = torch.device("cuda:0")
            elif torch.backends.mps.is_available():
                device = torch.device("mps")
            else:
                device = torch.device("cpu")
        except ImportError:
            raise WingmanException(
                "Could not import torch, this is not bundled as part of Wingman and has to be installed manually."
            )

        return device

    def checkpoint(
        self, loss: float, step: int | None = None
    ) -> tuple[bool, Path, Path]:
        """checkpoint.

        Records training every logging_interval steps.

        Returns three things:
        - indicator on whether we should save weights
        - path of where the weight files should be saved
        - path of where the optim files should be saved

        Args:
        ----
            loss (float): learning loss of the model as a detached float
            step (int | None): step number, automatically incremented if None

        Returns:
        -------
            tuple[bool, Path, Path]: to_update, weights_file, optim_file

        """
        # if step is None, we automatically increment
        if step is None:
            step = self._previous_ckpt_step + 1

        # check that our step didn't go in reverse
        if step < self._previous_ckpt_step:
            raise WingmanException(
                cstr(
                    f"We can't step backwards! Got step {step} but the previous logging step was {self._previous_ckpt_step}.",
                    "FAIL",
                )
            )
        self._previous_ckpt_step = step

        """ACCUMULATE LOSS"""
        self._cumulative_loss += loss
        self._num_losses += 1

        # if we haven't passed the required number of steps
        if step < self._next_log_step:
            return False, self._model_dir, self._ckpt_dir

        # log to wandb if needed, but only on the logging steps
        if self.cfg.wandb.enable:
            self.wandb_log()

        """GET NEW AVG LOSS"""
        # record the losses and reset the cumulative
        avg_loss = self._cumulative_loss / self._num_losses
        self._cumulative_loss = 0.0
        self._num_losses = 0

        # compute the next time we need to log
        self._next_log_step = (
            int(step / self.cfg.logging.interval) + 1
        ) * self.cfg.logging.interval

        # always print on n intervals
        wm_print(
            f"Step {cstr(step, 'OKCYAN')}; "
            f"Average Loss {cstr(f'{avg_loss:.5f}', 'OKCYAN')}; "
            f"Lowest Average Loss {cstr(f'{self._lowest_cumulative_lost:.5f}', 'OKCYAN')}",
            self.cfg.logging.filename,
        )

        """CHECK IF NEW LOSS IS BETTER"""
        # if we don't meet the criteria for saving
        if avg_loss >= self._lowest_cumulative_lost - self.cfg.logging.greater_than:
            # accumulate skips
            if self._skips < self.cfg.logging.max_skips:
                self._skips += 1
                return False, self._model_dir, self._ckpt_dir
            else:
                # save the network to intermediary if we crossed the max number of skips
                wm_print(
                    f"Passed {self.cfg.logging.max_skips} intervals without saving so far. "
                    f"Issuing new checkpoint directory: {cstr(self._temp_dir, 'OKCYAN')}",
                    self.cfg.logging.filename,
                )
                self._skips = 0
                return True, self._model_dir, self._temp_dir

        """NEW LOSS IS BETTER"""
        # redefine the new lowest loss and reset the skips
        self._lowest_cumulative_lost = avg_loss
        self._skips = 0

        # increment means return the files with incremented checkpoint number
        if self.cfg.model.increment_ckpt:
            # check if we are safe to increment checkpoint numbers and regenerate weights file
            if self._previous_ckpt == -1 or self._ckpt_dir.exists():
                self._previous_ckpt = self._current_ckpt
                self._current_ckpt += 1

            else:
                wm_print(
                    cstr(
                        f"Didn't populate previous checkpoint directory ({self._ckpt_dir}). "
                        "Therefore, not incrementing checkpoint number.",
                        "WARNING",
                    ),
                    self.cfg.logging.filename,
                )
                self._current_ckpt = self._previous_ckpt

        # record the lowest running loss in the status file
        np.save(self._lowest_loss_file, self._lowest_cumulative_lost)

        wm_print(
            f"New lowest point, issuing new checkpoint directory: {cstr(self._ckpt_dir, 'OKGREEN')}",
            self.cfg.logging.filename,
        )

        return True, self._model_dir, self._ckpt_dir

    def wandb_log(self) -> None:
        """wandb_log.

        Logs the internal log to WandB.
        Start logging by adding things to it using:
        ```
        wm.log["foo"] = "bar"
        ```

        Returns
        -------
            None:

        """
        if not isinstance(self.log, dict):
            raise WingmanException(
                cstr(f"log must be dictionary, currently it is {self.log}.", "FAIL")
            )
        if self.cfg.wandb.enable:
            wandb.log(self.log)

    def write_auxiliary(
        self, data: np.ndarray, variable_name: str, precision: str = "%1.3f"
    ) -> None:
        """write_auxiliary.

        Args:
        ----
            data (np.ndarray): data
            variable_name (str): variable_name
            precision (str): precision

        Returns:
        -------
            None:

        """
        if not len(data.shape) == 1:
            raise WingmanException(
                cstr("Data must be only 1 dimensional ndarray", "FAIL")
            )
        filename = self._model_dir / f"{variable_name}.csv"
        with open(filename, "ab") as f:
            np.savetxt(f, [data], delimiter=",", fmt=precision)

    def get_weight_files(self, latest: bool = True) -> tuple[bool, Path, Path]:
        """get_weight_files.

        Returns three things:
        - indicator on whether we have weight files
        - directory of where the weight files are
        - directory of where the optim files are

        Args:
        ----
            latest (bool): whether we want the latest file or the one determined by `ckpt_number`

        Returns:
        -------
            tuple[bool, Path, Path]: have_file, weights_file, optim_file

        """
        # if we don't need the latest file, get the one specified
        if not latest:
            ckpt_dir = self._model_dir / f"{self._current_ckpt}/"
            if ckpt_dir.exists():
                wm_print(
                    f"Using checkpoint directory: {cstr(f'{self._ckpt_dir}', 'OKGREEN')}",
                    self.cfg.logging.filename,
                )
                return True, self._model_dir, self._ckpt_dir
            else:
                raise ValueError(
                    cstr(
                        f"Checkpoint number {self._current_ckpt} was requested, "
                        f"but checkpoint directory {ckpt_dir} does not exist.",
                        "FAIL",
                    )
                )

        # while the file exists, try to look for a file one checkpoint later
        # once the checkpoint doesn't exist, decrement by one and use that file
        self._current_ckpt = 0
        while (self._model_dir / f"{self._current_ckpt}/").exists():
            self._current_ckpt += 1
        self._current_ckpt = max(self._current_ckpt - 1, 0)

        # if the file doesn't exist, notify and ignore
        if not (self._model_dir / f"{self._current_ckpt}/").exists():
            if not self._fresh_directory:
                wm_print(
                    cstr(
                        "No checkpoint directory found, generating new one during training.",
                        "WARNING",
                    ),
                    self.cfg.logging.filename,
                )
            self._fresh_directory = False

            return False, self._model_dir, self._ckpt_dir
        else:
            # hitch a ride to update the lowest running loss
            self._lowest_cumulative_lost = np.load(self._lowest_loss_file).item()

            wm_print(
                f"Using checkpoint directory: {cstr(f'{self._ckpt_dir}', 'OKGREEN')}",
                self.cfg.logging.filename,
            )
            wm_print(
                f"Lowest Running Loss: {cstr(self._lowest_cumulative_lost, 'OKCYAN')}",
                self.cfg.logging.filename,
            )

            return True, self._model_dir, self._ckpt_dir



"""Easy creation of neural networks."""

from typing import List, Optional

from wingman.exceptions import NeuralBlocksException

try:
    import torch.nn as nn
except ImportError as e:
    raise ImportError(
        "Could not import torch, this is not bundled as part of Wingman and has to be installed manually."
    ) from e


class NeuralBlocks:
    """NeuralBlocks."""

    def __init__(self):
        """__init__."""
        pass

    @classmethod
    def conv_module(
        cls,
        in_channel: int,
        out_channel: int,
        kernel_size: int,
        pooling: int,
        activation: str,
        pool_method: str = "max",
        padding: Optional[int] = None,
        norm: str = "non",
    ) -> nn.Sequential:
        """conv_module.

        Args:
        ----
            in_channel (int): in_channel
            out_channel (int): out_channel
            kernel_size (int): kernel_size
            pooling (int): pooling
            activation (str): activation
            pool_method (str): pool_method
            padding (Optional[int]): padding
            norm (str): norm

        Returns:
        -------
            nn.Sequential:

        """
        module_list = []

        # batch norm
        if norm != "non":
            module_list.append(
                cls.get_normalization(norm, num_features=in_channel, dimension=2)
            )

        # conv module
        if padding is not None:
            module_list.append(
                nn.Conv2d(
                    in_channels=in_channel,
                    out_channels=out_channel,
                    kernel_size=kernel_size,
                    padding=padding,
                )
            )
        else:
            module_list.append(
                nn.Conv2d(
                    in_channels=in_channel,
                    out_channels=out_channel,
                    kernel_size=kernel_size,
                    padding=kernel_size // 2,
                )
            )

        # if pooling then add pooling
        if pooling > 0:
            if pool_method == "max":
                module_list.append(nn.MaxPool2d(kernel_size=pooling))
            else:
                module_list.append(nn.AvgPool2d(kernel_size=pooling))

        # add in the activation function
        module_list.append(cls.get_activation(activation))

        return nn.Sequential(*module_list)

    @classmethod
    def linear_module(
        cls,
        in_features: int,
        out_features: int,
        activation: str,
        norm: str = "non",
        bias: bool = True,
    ) -> nn.Sequential:
        """linear_module.

        Args:
        ----
            in_features (int): in_features
            out_features (int): out_features
            activation (str): activation
            norm (str): norm
            bias (bool): bias

        Returns:
        -------
            nn.Sequential:

        """
        module_list = []

        # batch norm
        if norm != "non":
            module_list.append(
                cls.get_normalization(norm, num_features=in_features, dimension=1)
            )

        # linear module
        module_list.append(nn.Linear(in_features, out_features, bias))

        # add in the activation function
        module_list.append(cls.get_activation(activation))

        return nn.Sequential(*module_list)

    @classmethod
    def deconv_module(
        cls,
        in_channel: int,
        out_channel: int,
        kernel_size: int,
        padding: int,
        stride: int,
        activation: str,
        norm: str = "non",
    ) -> nn.Sequential:
        """deconv_module.

        Args:
        ----
            in_channel (int): in_channel
            out_channel (int): out_channel
            kernel_size (int): kernel_size
            padding (int): padding
            stride (int): stride
            activation (str): activation
            norm (str): norm

        Returns:
        -------
            nn.Sequential:

        """
        module_list = []

        # batch norm
        if norm != "non":
            module_list.append(
                cls.get_normalization(norm, num_features=in_channel, dimension=2)
            )

        # conv module
        module_list.append(
            nn.ConvTranspose2d(
                in_channels=in_channel,
                out_channels=out_channel,
                kernel_size=kernel_size,
                padding=padding,
                stride=stride,
            )
        )

        # add in the activation function
        module_list.append(cls.get_activation(activation))

        return nn.Sequential(*module_list)

    @classmethod
    def get_activation(cls, activation: str) -> nn.Module:
        """get_activation.

        Args:
        ----
            activation (str): activation

        Returns:
        -------
            nn.Module:

        """
        if activation == "sigmoid":
            return nn.Sigmoid()
        elif activation == "relu":
            return nn.ReLU()
        elif activation == "lrelu":
            return nn.LeakyReLU()
        elif activation == "tanh":
            return nn.Tanh()
        elif activation == "identity":
            return nn.Identity()
        else:
            raise NotImplementedError

    @classmethod
    def get_normalization(
        cls, activation: str, num_features: int, dimension: int
    ) -> nn.Module:
        """get_normalization.

        Args:
        ----
            activation (str): activation
            num_features (int): num_features
            dimension (int): dimension

        Returns:
        -------
            nn.Module:

        """
        if activation == "batch":
            if dimension == 1:
                return nn.BatchNorm1d(num_features)
            elif dimension == 2:
                return nn.BatchNorm2d(num_features)
            else:
                raise NotImplementedError
        elif activation == "layer":
            if dimension == 1:
                return nn.LayerNorm(num_features)
            else:
                raise NotImplementedError
        else:
            raise NotImplementedError

    @classmethod
    def generate_conv_stack(
        cls,
        channels_description: List[int],
        kernels_description: List[int],
        pooling_description: List[int],
        activation_description: List[str],
        padding: Optional[int] = None,
        norm: str = "non",
    ) -> nn.Sequential:
        """generate_conv_stack.

        Args:
        ----
            channels_description (List[int]): channels_description
            kernels_description (List[int]): kernels_description
            pooling_description (List[int]): pooling_description
            activation_description (List[str]): activation_description
            padding (Optional[int]): padding
            norm (str): norm

        Returns:
        -------
            nn.Sequential:

        """
        network_depth = len(channels_description) - 1

        if (
            network_depth != len(kernels_description)
            or network_depth != len(activation_description)
            or network_depth != len(pooling_description)
        ):
            raise NeuralBlocksException(
                "All network descriptions must be of the same size"
            )

        module_list = []
        for i in range(network_depth):
            module_list.append(
                cls.conv_module(
                    channels_description[i],
                    channels_description[i + 1],
                    kernels_description[i],
                    pooling_description[i],
                    activation_description[i],
                    padding=padding,
                    norm=norm,
                )
            )

        return nn.Sequential(*module_list)

    @classmethod
    def generate_deconv_stack(
        cls,
        channels_description: List[int],
        kernels_description: List[int],
        padding_description: List[int],
        stride_description: List[int],
        activation_description: List[str],
        norm: str = "non",
    ) -> nn.Sequential:
        """generate_deconv_stack.

        Args:
        ----
            channels_description (List[int]): channels_description
            kernels_description (List[int]): kernels_description
            padding_description (List[int]): padding_description
            stride_description (List[int]): stride_description
            activation_description (List[str]): activation_description
            norm (str): norm

        Returns:
        -------
            nn.Sequential:

        """
        network_depth = len(channels_description) - 1

        if (
            network_depth != len(kernels_description)
            or network_depth != len(activation_description)
            or network_depth != len(padding_description)
            or network_depth != len(stride_description)
        ):
            raise NeuralBlocksException(
                "All network descriptions must be of the same size"
            )

        module_list = []
        for i in range(network_depth):
            module_list.append(
                cls.deconv_module(
                    channels_description[i],
                    channels_description[i + 1],
                    kernels_description[i],
                    padding_description[i],
                    stride_description[i],
                    activation_description[i],
                    norm=norm,
                )
            )

        return nn.Sequential(*module_list)

    @classmethod
    def generate_linear_stack(
        cls,
        features_description: List[int],
        activation_description: List[str],
        norm: str = "non",
        bias: bool = True,
    ) -> nn.Sequential:
        """generate_linear_stack.

        Args:
        ----
            features_description (List[int]): features_description
            activation_description (List[str]): activation_description
            norm (str): norm
            bias (bool): bias

        Returns:
        -------
            nn.Sequential:

        """
        network_depth = len(features_description) - 1

        if network_depth != len(activation_description):
            raise NeuralBlocksException(
                f"All network descriptions must be of the same size, got {network_depth} for `network_depth` and {activation_description} for `activation_description`."
            )

        module_list = []
        for i in range(network_depth):
            module_list.append(
                cls.linear_module(
                    features_description[i],
                    features_description[i + 1],
                    activation_description[i],
                    norm=norm,
                    bias=bias,
                )
            )

        return nn.Sequential(*module_list)
